import errno
import logging
import os
import Queue
import signal
import sys
import time

from gevent import pool, monkey, select


class Worker(object):

    understood_signals = ["QUIT", "TERM", "INT"]

    def __init__(self, in_queue, taken_queue, out_queue, config):
        self.in_queue = in_queue
        self.taken_queue = taken_queue
        self.out_queue = out_queue
        self.config = config

        self.alive = True
        self.active = True
        self.age = 0

        self.smtp_pool = None

        self.logger = logging.getLogger(__name__)

    def setup(self):
        self.pid = os.getpid()
        self.parent_pid = os.getppid()
        self.setup_signals()

    def setup_signals(self):

        for signal_name in self.understood_signals:
            signal.signal(
                getattr(signal, "SIG%s" % signal_name),
                getattr(self, "handle_%s" % signal_name.lower())
            )

    def run(self):
        monkey.patch_all(thread=False)

        greenlet_pool = pool.Pool(self.config.greenlet_pool_size)

        while self.alive:
            try:
                ready = select.select(
                    [self.in_queue._reader],
                    [], [],
                    self.config.heartbeat_interval
                )
                if not ready[0] or not self.active:
                    self.heartbeat()

                self.age += 1
                job = self.in_queue.get_nowait()
                greenlet_pool.spawn(
                    self.handle_job, job
                )

            except Queue.Empty:
                continue
            except select.error as error:
                if error.args[0] not in [errno.EAGAIN, errno.EINTR]:
                    raise
            except OSError as e:
                if e.errno not in [errno.EAGAIN, errno.EINTR]:
                    raise
            except SystemExit:
                raise
            except Exception as e:
                self.logger.exception(e)
                sys.exit(-1)

        greenlet_pool.join(raise_error=True)

    def heartbeat(self):
        if self.pid == os.getpid() and self.parent_pid != os.getppid():
            self.logger.info("Parent process changed! shutting down.")
            self.alive = False

    def handle_job(self, job):
        start_time = time.time()
        self.taken_queue.put(
            {
                "time": start_time,
                "job": job
            }
        )

        try:
            job.load()
        except:
            self.logger.exception("Exception when loading job!")
        else:
            try:
                job.run()
            except:
                self.logger.exception("Exception when running job!")

        end_time = time.time()

        self.out_queue.put(
            {
                "time": end_time - start_time,
                "job": job
            }
        )

    def handle_int(self, signal, frame):
        self.active = not self.active

    def handle_quit(self, signal, frame):
        self.active = False
        self.alive = False

    def handle_term(self, signal, frame):
        self.active = False
        self.alive = False
        sys.exit(0)



import ConfigParser


class Config(object):

    def __init__(self, filename):
        self.filename = filename

    def load(self):
        parser = ConfigParser.SafeConfigParser()
        parser.read(self.filename)

        for section_name in parser.sections():
            setattr(self, section_name, Section())

            for option_name, option_value in parser.items(section_name):
                if "." in option_value:
                    try:
                        option_value = parser.getfloat(
                            section_name, option_name
                        )
                        setattr(
                            getattr(self, section_name),
                            option_name, option_value
                        )
                        continue
                    except ValueError:
                        pass
                else:
                    try:
                        option_value = parser.getint(
                            section_name, option_name
                        )
                        setattr(
                            getattr(self, section_name),
                            option_name, option_value
                        )
                        continue
                    except ValueError:
                        pass
                try:
                    option_value = parser.getboolean(
                        section_name, option_name
                    )
                    setattr(
                        getattr(self, section_name),
                        option_name, option_value
                    )
                    continue
                except ValueError:
                    pass

                setattr(
                    getattr(self, section_name),
                    option_name, option_value
                )


class Section(object):
    pass



import os
import time
import types


def get_script_content(command):
    current_path = os.path.dirname(__file__)

    content = ""
    with open("%s/lua/%s.lua" % (current_path, command)) as fd:
        content = fd.read()

    return content


def add_qadd(client):
    content = get_script_content("qadd")

    method = client.register_script(content)

    def qadd(self, queue, when, job_key, job_payload):
        return method(
            keys=[
                queue + ":schedule",
                queue + ":working",
                queue + ":job_pool"
            ],
            args=[when, job_key, job_payload],
            client=self
        )

    client.qadd = types.MethodType(qadd, client)


def add_qget(client):
    content = get_script_content("qget")

    method = client.register_script(content)

    def qget(self, queue, cutoff, start=0):
        return method(
            keys=[
                queue + ":schedule",
                queue + ":job_pool"
            ],
            args=[start, cutoff],
            client=self
        )

    client.qget = types.MethodType(qget, client)


def add_qsetstate(client):
    content = get_script_content("qsetstate")

    method = client.register_script(content)

    def qsetstate(self, queue, current, target, *job_keys):
        args = [time.time()]
        args.extend(job_keys)

        return method(
            keys=[
                queue + ":" + current,
                queue + ":" + target
            ],
            args=args,
            client=self
        )

    client.qsetstate = types.MethodType(qsetstate, client)


def extend_redis(client):
    add_qadd(client)
    add_qget(client)
    add_qsetstate(client)


__all__ = [extend_redis]



import errno
import logging
import os
import socket

from .job import Job


SOCKET_BUFFER_SIZE = 4096


class Connection(object):

    def __init__(self, host='', port=8765):
        self.host = host
        self.port = port

        self.socket = None

        self.logger = logging.getLogger(__name__)

    def open(self):
        existing_fd = None
        if "ROTTERDAM_SOCKET_FD" in os.environ:
            existing_fd = int(os.environ.pop("ROTTERDAM_SOCKET_FD"))
            self.socket = socket.fromfd(
                existing_fd,
                socket.AF_INET, socket.SOCK_STREAM
            )
        else:
            self.socket = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM
            )
        self.socket.setblocking(0)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if existing_fd is None:
            self.socket.bind((self.host, self.port))
        self.socket.listen(5)

    def close(self):
        self.socket.close()

    def iterjobs(self):
        accepted = False

        while not accepted:
            try:
                conn, addr = self.socket.accept()
                accepted = True
            except socket.error as e:
                if e.errno == errno.EAGAIN:
                    accepted = False

        self.logger.debug("connection from %s:%s", addr[0], addr[1])

        message = ''

        while True:
            try:
                chunk = conn.recv(SOCKET_BUFFER_SIZE)
                if chunk:
                    message = message + chunk
            except socket.error as e:
                if e.errno not in [errno.EAGAIN, errno.EINTR]:
                    raise

            if not message:
                break

            if "\n" in message:
                message, extra = message.split("\n", 1)

            job = Job()
            job.deserialize(message)

            yield job

            message = extra

        conn.close()



import hashlib
import json
import types

from .exceptions import NoSuchJob


class Job(object):

    def __init__(self):
        self.when = None
        self.unique_key = None

        self.module = None
        self.function = None
        self.args = None
        self.kwargs = None

        self.call = None

    def from_function(self, func, args, kwargs):
        self.args = args
        self.kwargs = kwargs

        if isinstance(func, basestring):
            self.module, self.function = func.split(":")
        elif isinstance(func, types.FunctionType):
            self.module = func.__module__
            self.function = func.__name__

        uniqueness = hashlib.md5()

        uniqueness.update(str(self.module))
        uniqueness.update(str(self.function))

        for arg in args:
            uniqueness.update(str(arg))
        for arg_name in sorted(kwargs.keys()):
            uniqueness.update(
                str(arg_name) + "=" + str(kwargs[arg_name])
            )

        self.unique_key = uniqueness.hexdigest()

    def serialize(self):
        return json.dumps({
            "when": self.when,
            "unique_key": self.unique_key,
            "module": self.module,
            "function": self.function,
            "args": self.args,
            "kwargs": self.kwargs
        })

    def deserialize(self, payload):
        payload = json.loads(payload)

        for attribute in [
            'when', 'unique_key', 'module', 'function', 'args', 'kwargs'
        ]:
            setattr(self, attribute, payload[attribute])

    def load(self):
        try:
            module = __import__(self.module, fromlist=self.function)
            self.call = getattr(module, self.function)
        except ImportError:
            raise NoSuchJob

    def run(self):
        return self.call(*self.args, **self.kwargs)

    def __repr__(self):
        arg_string = ", ".join(self.args)
        kwarg_string = ", ".join([
            "%s=%s" % (name, val)
            for name, val in self.kwargs.iteritems()
        ])
        if arg_string and kwarg_string:
            kwarg_string = ", " + kwarg_string
        return "%(module)s.%(func_name)s(%(args)s%(kwargs)s)" % {
            "module": self.module,
            "func_name": self.function,
            "args": arg_string,
            "kwargs": kwarg_string
        }



import errno
import signal
import os

from .config import Config


class Control(object):

    def __init__(self, args):
        self.args = args
        self.pid = None

    def load_config(self):
        self.config = Config(self.args.config_file)
        self.config.load()

    def get_pid(self):
        try:
            pid_file = open(self.config.master.pid_file, 'r')
        except IOError as e:
            if e.errno == errno.ENOENT:
                raise RuntimeError("Mailer not currently running!")
            else:
                raise

        self.pid = int(pid_file.read())

        pid_file.close()

    def run(self):
        self.load_config()
        self.get_pid()

        getattr(self, self.args.command)()

    def stop(self):
        os.kill(self.pid, signal.SIGQUIT)

    def reload(self):
        os.kill(self.pid, signal.SIGHUP)

    def relaunch(self):
        os.kill(self.pid, signal.SIGUSR1)

    def expand(self):
        os.kill(self.pid, signal.SIGTTIN)

    def contract(self):
        os.kill(self.pid, signal.SIGTTOU)



import errno
import logging
import Queue
import select
import time

import redis

from .job import Job
from .exceptions import NoSuchJob
from .redis_extensions import extend_redis


class Arbiter(object):

    def __init__(
            self,
            config,
            connection,
            ready_queue, taken_queue, results_queue
    ):
        self.config = config
        self.connection = connection

        self.ready_queue = ready_queue
        self.taken_queue = taken_queue
        self.results_queue = results_queue

        self.logger = logging.getLogger(__name__)

        self.redis = None

    def setup(self):

        if ":" in self.config.redis_host:
            host, port = self.config.redis_host.split(":")
            self.redis = redis.Redis(host=host, port=port)
        else:
            self.redis = redis.Redis(host=self.config.redis_host)

        extend_redis(self.redis)

        self.fill_ready_queue()

    def arbitrate(self, timeout):
        try:
            (sources_with_data, [], []) = select.select(
                [
                    self.connection.socket,
                    self.taken_queue._reader,
                    self.results_queue._reader
                ],
                [], [],
                timeout
            )

            while len(sources_with_data) > 0:
                source_with_data = sources_with_data.pop(0)

                if source_with_data == self.connection.socket:
                    self.handle_incoming_job()
                elif source_with_data == self.taken_queue._reader:
                    self.handle_taken_job()
                elif source_with_data == self.results_queue._reader:
                    self.handle_finished_job()

        except select.error as e:
            if e.args[0] not in [errno.EAGAIN, errno.EINTR]:
                raise
        except OSError as e:
            if e.errno not in [errno.EAGAIN, errno.EINTR]:
                raise
        except IOError as e:
            if e.errno not in [errno.EAGAIN, errno.EINTR, errno.EBADF]:
                raise

    def fill_ready_queue(self):
        while True:
            try:
                payloads = self.redis.qget("rotterdam", int(time.time()))

                if not payloads:
                    break

                for payload in payloads:
                    job = Job()
                    job.deserialize(payload)
                    self.logger.debug(
                        "Queueing job: %s", job
                    )
                    self.redis.qsetstate(
                        "rotterdam",
                        "schedule",
                        "ready",
                        job.unique_key
                    )
                    self.ready_queue.put_nowait(job)
            except Queue.Full:
                break

    def handle_incoming_job(self):
        for job in self.connection.iterjobs():
            self.logger.debug("got job: %s", job)
            try:
                job.load()
            except NoSuchJob:
                continue

            self.redis.qadd(
                "rotterdam",
                job.when,
                job.unique_key,
                job.serialize()
            )

        self.fill_ready_queue()

    def handle_taken_job(self):
        while True:
            try:
                taken = self.taken_queue.get_nowait()

                self.logger.debug("Job started %s", taken["job"])

                self.redis.qsetstate(
                    "rotterdam",
                    "ready",
                    "working",
                    taken["job"].unique_key
                )

            except Queue.Empty:
                break

        self.fill_ready_queue()

    def handle_finished_job(self):
        while True:
            try:
                result = self.results_queue.get_nowait()

                self.logger.debug(
                    "Job completed in %0.2fs seconds",
                    result["time"]
                )

                self.redis.qsetstate(
                    "rotterdam",
                    "working",
                    "done",
                    result["job"].unique_key
                )

            except Queue.Empty:
                break



import logging
import socket
import time

from .exceptions import ConnectionError
from .job import Job

SOCKET_BUFFER_SIZE = 4096


class Client(object):

    def __init__(self, host, port=8765):
        self.host = host
        self.port = port

        self.socket = None

        self.logger = logging.getLogger(__name__)

    @property
    def connected(self):
        return self.socket and self.socket.fileno()

    def connect(self):
        if self.connected:
            return

        self.logger.debug("connecting to %s:%s", self.host, self.port)

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.host, self.port))

    def enqueue(self, func, *args, **kwargs):
        job = Job()
        job.from_function(func, args, kwargs)
        job.when = int(time.time())

        self.logger.debug("enqueuing job %s", job)

        self.connect()

        try:
            self.socket.sendall(job.serialize() + "\n")
        except IOError, e:
            raise ConnectionError(
                "Error sending job to %s:%s, %s" % (
                    self.host,
                    self.port,
                    e.args[1] if len(e.args) > 1 else e.args[0]
                )
            )
        finally:
            self.disconnect()

    def disconnect(self):
        try:
            self.logger.debug("disconnecting from %s:%s", self.host, self.port)
            self.socket.close()
        except socket.error:
            pass
        self.socket = None

    def __del__(self):
        if self.connected:
            self.disconnect()



from setuptools import setup

setup(
    name="rotterdam",
    version="0.0.5",
    packages=["rotterdam"],
    package_data={
        'rotterdam': ['lua/*.lua']
    },
    scripts=[
        "bin/rotterdam",
        "bin/rotterdamctl"
    ],
    install_requires=[
        "cython==0.19.1",
        "setproctitle",
        "gevent>=1.0dev",
        "redis==2.7.6"
    ],
    dependency_links=[
        "http://github.com/surfly/gevent/tarball/1.0rc3#egg=gevent-1.0dev"
    ],
    tests_require=[
        "mock==1.0.1",
        "nose==1.3.0"
    ]
)



import errno
import logging
import os
import signal
import sys
import time
from multiprocessing import queues

import setproctitle

from .config import Config
from .connection import Connection
from .worker import Worker
from .arbiter import Arbiter


class Master(object):

    signal_map = {
        "hup": "reload_config",
        "usr1": "relaunch",
        "ttin": "expand_workers",
        "ttou": "contract_workers",
        "int": "halt_current_jobs",
        "quit": "wind_down_gracefully",
        "term": "wind_down_immediately",
        "chld": "handle_worker_exit"
    }

    def __init__(self, config_file):
        self.pid = None
        self.pid_file_path = None
        self.reexec_pid = 0

        self.name = "master"

        self.logger = logging.getLogger(__name__)

        self.config_file = config_file

        self.workers = {}

        self.connection = None

        self.ready_queue = None
        self.taken_queue = None
        self.results_queue = None

        self.wind_down_time = None

        args = sys.argv[:]
        args.insert(0, sys.executable)

        self.launch_context = {
            "executable": sys.executable,
            "args": args,
            "cwd": os.getcwd()
        }

    def load_config(self):
        self.config = Config(self.config_file)
        self.config.load()
        self.pid_file_path = self.config.master.pid_file
        self.number_of_workers = self.config.master.num_workers

    def setup_pid_file(self):
        if not os.path.isdir(os.path.dirname(self.pid_file_path)):
            raise RuntimeError(
                "%s does not exist! Can't create pid file" % self.pid_file_path
            )

        if os.path.exists(self.pid_file_path):
            raise RuntimeError(
                (
                    "pid file (%s) already exists! \n"
                    "If no master is running it could be stale."
                ) % self.pid_file_path
            )

        fd = open(self.pid_file_path, 'w')
        try:
            fd.write(u"%s\n" % self.pid)
        except:
            os.unlink(self.pid_file_path)
            raise
        else:
            fd.close()

    def setup_connection(self):
        self.connection = Connection(
            port=self.config.master.listen_port
        )
        self.connection.open()
        self.logger.info(
            "Listening on port %s", self.config.master.listen_port
        )

    def setup_signals(self):
        for signal_name, handler_name in self.signal_map.iteritems():
            signal.signal(
                getattr(signal, "SIG%s" % signal_name.upper()),
                getattr(self, handler_name)
            )

    def setup_ipc_queues(self):
        self.ready_queue = queues.Queue(
            maxsize=(
                self.number_of_workers * self.config.workers.greenlet_pool_size
            )
        )
        self.taken_queue = queues.Queue()
        self.results_queue = queues.Queue()

    def run(self):
        self.pid = os.getpid()
        self.logger.info("Starting %s (%d)", self.name, int(self.pid))

        self.load_config()
        self.setup_pid_file()
        self.setup_ipc_queues()
        self.setup_connection()
        self.setup_signals()

        setproctitle.setproctitle("rotterdam: %s" % self.name)

        for i in range(self.number_of_workers):
            self.spawn_worker()

        arbiter = Arbiter(
            self.config.arbiter,
            self.connection,
            self.ready_queue,
            self.taken_queue,
            self.results_queue
        )

        arbiter.setup()

        while True:
            try:
                arbiter.arbitrate(
                    timeout=self.config.master.heartbeat_interval
                )
                self.heartbeat()
            except SystemExit:
                raise
            except KeyboardInterrupt:
                self.wind_down_gracefully()
            except Exception, e:
                self.logger.exception(e)
                sys.exit(-1)

    def expand_workers(self, *args):
        self.number_of_workers += 1
        self.logger.info(
            "Upping number of workers to %d",
            self.number_of_workers
        )
        self.spawn_worker()

    def contract_workers(self, *args):
        if self.number_of_workers <= 1:
            self.logger.info(
                "Ignoring TTOU, number of workers already at %d",
                self.number_of_workers
            )
            return

        self.number_of_workers -= 1
        self.logger.info(
            "Contracting number of workers to %d",
            self.number_of_workers
        )

        # get the pid of the oldest worker by
        # sorting by age and popping the first one
        oldest_worker_pid, _ = sorted(
            self.workers.items(),
            key=lambda i: i[1].age
        ).pop(0)

        self.send_signal(signal.SIGQUIT, oldest_worker_pid)

    def reload_config(self, *args):
        self.logger.info("Reloading config")
        old_port = self.config.master.listen_port

        workers_by_age = sorted(
            self.workers.items(),
            key=lambda i: i[1].age
        )

        self.load_config()

        if self.config.master.listen_port != old_port:
            self.connection.close()
            self.setup_connection()

        for i in range(self.number_of_workers):
            self.spawn_worker()

        while len(workers_by_age) > 0:
            (worker_pid, _) = workers_by_age.pop(0)
            self.send_signal(signal.SIGQUIT, worker_pid)

    def relaunch(self, *args):
        os.rename(
            self.pid_file_path,
            self.pid_file_path + ".old." + str(self.pid)
        )

        self.reexec_pid = os.fork()

        if self.reexec_pid != 0:
            # old master proc
            self.name = "old master"
            self.pid_file_path += ".old." + str(self.pid)
            setproctitle.setproctitle("rotterdam: %s" % self.name)
            self.wind_down()
            return

        os.environ["ROTTERDAM_SOCKET_FD"] = str(
            self.connection.socket.fileno()
        )

        os.chdir(self.launch_context["cwd"])

        os.execvpe(
            self.launch_context["executable"],
            self.launch_context["args"],
            os.environ
        )

    def handle_worker_exit(self, *args):
        worker_pids = self.workers.keys()
        for worker_pid in worker_pids:
            try:
                pid, status = os.waitpid(worker_pid, os.WNOHANG)
                if pid == worker_pid:
                    self.workers.pop(worker_pid)
            except OSError as e:
                if e.errno == errno.ECHILD:
                    self.workers.pop(worker_pid)
                raise

        if self.wind_down_time:
            if not self.workers:
                try:
                    os.unlink(self.pid_file_path)
                except OSError as e:
                    if e.errno == errno.ENOENT:
                        pass
                sys.exit(0)
            else:
                time.sleep(self.wind_down_time - time.time())
                self.broadcast_signal(signal.SIGKILL)

    def halt_current_jobs(self, *args):
        self.broadcast_signal(signal.SIGINT)

    def wind_down_gracefully(self, *args):
        self.connection.close()
        self.wind_down()

    def wind_down_immediately(self, *args):
        self.connection.close()
        self.wind_down(signal_to_broadcast=signal.SIGTERM)

    def spawn_worker(self):
        worker = Worker(
            self.ready_queue,
            self.taken_queue,
            self.results_queue,
            self.config.workers
        )

        pid = os.fork()

        if pid != 0:
            self.workers[pid] = worker
            return

        # in the worker process now
        try:
            setproctitle.setproctitle("rotterdam: worker")
            self.logger.info("Starting up worker")
            worker.setup()
            worker.run()
            sys.exit(0)
        except SystemExit:
            raise
        except:
            self.logger.exception("Exception in worker process!")
            sys.exit(-1)
        finally:
            self.logger.info("Worker exiting")

    def wind_down(self, signal_to_broadcast=signal.SIGQUIT):
        self.logger.info("Winding down %s", self.name)
        self.wind_down_time = (
            time.time() + self.config.master.shutdown_grace_period
        )
        self.broadcast_signal(signal_to_broadcast)

    def broadcast_signal(self, signal):
        for worker_pid in self.workers.keys():
            self.send_signal(signal, worker_pid)

    def send_signal(self, signal, worker_pid):
        try:
            os.kill(worker_pid, signal)
        except OSError as error:
            if error.errno == errno.ESRCH:
                try:
                    self.workers.pop(worker_pid)
                except KeyError:
                    return
            raise

    def heartbeat(self):
        pass
